/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <gpxe/malloc.h>
#include <gpxe/command.h>
#include <gpxe/settings.h>
#include <gpxe/sf_settings.h>
#include <gpxe/sanity_checks.h>
#include <gpxe/keys.h>
#include <gpxe/process.h>
#include <gpxe/timer.h>
#include <gpxe/netdevice.h>
#include <gpxe/device.h>
#include <gpxe/pci.h>
#include <gpxe/dhcp.h>
#include <gpxe/iscsi.h>
#include <gpxe/process.h>
#include <gpxe/sf_widgets.h>
#include <gpxe/sf_main.h>
#include <gpxe/nvo.h>
#include <gpxe/dyncfg.h>
#include "mc_driver_pcol.h"
#include "tlv_layout.h"
#include "pci_ids.h"

#include "sf_brwidgets.h"
#include "../../src/drivers/net/sfc/ef10/dyncfg_accessors.h"

#define ARRAY_SIZE( _x ) ( sizeof ( _x ) / sizeof ( _x[0] ) )

#define for_each_widget(w, win) \
	for ((w) = next_widget(win);(w) && (w) != (win); (w) = next_widget(w))
#define for_each_screen_widget(w, conf, screen) \
	for_each_widget(w, (conf)->screens[screen])

int reboot_needed = 0;

/* Global screens */
#ifdef BCV_PROTO_ISCSI
static sf_widget *mpio_scr = NULL;
static sf_widget *mpio_scr_list = NULL;
#endif
static sf_widget *main_scr = NULL;
static sf_widget *main_scr_summary = NULL;
static sf_widget *exit_scr = NULL;

struct netdev_conf *all_netdevs = NULL;

/* Forward declarations */
static void refresh_pf_list ( struct netdev_conf *conf );
#ifdef BCV_PROTO_ISCSI
static void reorder_mpio ( void );
#endif

/* Returns the pci_device struct of a net device.
 * Unsafe if the device is not PCI.
 */
static struct pci_device *sf_netdev_pcidev ( struct net_device *dev )
{
	return (struct pci_device *) dev->dev;
}

static int is_ef10_device(uint16_t device)
{
	return SFU_DEVICE_TYPE(device) >= SFC_FARMINGDALE_PF_DEVICE_ID;
}

static bool conf_has_link_control(struct netdev_conf *conf)
{
	return conf->link_ctrl_conf == conf;
}

static void set_widgets_selectable ( sf_widget *w, int count, bool enable )
{
	while ( count-- ) {
		w->selectable = enable;
		sf_widget_draw ( w );
		w = next_widget ( w );
	}
}

static void set_fw_variant(sf_widget *this, int key)
{
	struct netdev_conf *pf_conf, *conf = this->aux_data;
	struct pci_device *pci_dev = sf_netdev_pcidev(conf->netdev);
	uint8_t this_pci_bus = pci_dev->bus;
	unsigned this_pci_slot = PCI_SLOT(pci_dev->devfn);
	sf_combo *variant;

	for_each_conf(pf_conf) {
		/* Only update this adapter */
		pci_dev = sf_netdev_pcidev(pf_conf->netdev);
		if (pci_dev->bus != this_pci_bus ||
		    PCI_SLOT(pci_dev->devfn) != this_pci_slot)
			continue;

		variant = pf_conf->fw_variant;
		if (!variant)
			continue;

		sf_combo_set(variant, key);
		/* The firmware variant is read-only, so by default
		 * sf_widget_save() will not save it.
		 */
		firmware_variant_setting.type = &setting_type_int32;
	}
}

static int sf_editbox_cb_check ( sf_widget *this )
{
	struct setting *setting = this->setting;
	const char *buf = sf_editbox_get ( SF_EDITBOX ( this ) );

	if ( buf[0] == '\0' )
		goto out;

	if ( setting->type == &setting_type_ipv4 ) {
		struct in_addr tmp;

		if ( inet_aton ( buf, &tmp ) == 0 )
			return RC_INVAL;

	} else if ( setting->type == &setting_type_uint8 ||
#ifdef BCV_PROTO_ISCSI
		    setting->type == &setting_type_lun ||
#endif
		    setting->type == &setting_type_uint16 ||
		    setting->type == &setting_type_uint32) {
		unsigned long value, max;
		char *end_ptr;

		value = strtoul ( buf, &end_ptr, 10 );
		if ( *end_ptr )
			return RC_INVAL;
		if (setting == &vlan_tag_setting )
			max = 4094; /* 4095 is reserved and not to be configured */
		else if ( setting->type == &setting_type_uint32 )
			max = 0xffffffff;
		else if ( setting->type == &setting_type_uint16 )
			max = 0xffff;
		else
			max = 0xff;
		if ( value > max )
			return RC_INVAL;

		if (setting == &vlan_tag_setting) {
			static const char *lines[] =
				{ "The switch mode will now be changed ",
				  "to partitioning and the firmware variant ",
				  "will now be changed to full-featured.",
				  "This is to avoid configuration mismatch." };

			set_fw_variant(this,
				       TLV_FIRMWARE_VARIANT_FULL_FEATURED);
			if ( sf_dialog ( NULL, "Changes to other settings",
					 SFW_DIALOG_OK, -1, -1, lines,
					 ARRAY_SIZE ( lines ) ) != 0 )
				return 0;
		}
	}

out:
	return 0;
}

#ifdef BCV_PROTO_ISCSI
static int sf_editbox_cb_check_secret ( sf_widget *this )
{
	const char *buf = sf_editbox_get ( SF_EDITBOX ( this ) );
	unsigned int len = strlen ( buf );

	return ( len == 0 || ( len >= 12 && len <= 20 ) ? 0 : RC_INVAL );
}

static int __pure sf_conf_is_iscsi ( struct netdev_conf *conf )
{
	int boot_image = conf->link_ctrl_conf->_bootimage->index;

	return ( ( boot_image == BOOTIMAGE_PXE_UEFI ||
		   boot_image == BOOTIMAGE_PXE ) &&
		 conf->boottype->index == BOOTTYPE_ISCSI );
}
#endif

/* This is the default button callback - display the next window list */
static int sf_button_cb ( sf_widget *this ) {
	struct netdev_conf	*conf = this->aux_data;

#ifdef BCV_PROTO_ISCSI
	if (this->aux_data2 == _ISCSI_MPIO)
		return sf_widget_run ( mpio_scr, 0 );
#endif

	return sf_widget_run ( conf->screens[this->aux_data2], 0 );
}

#define BOOT_IMAGE_OPTIONS \
	{ .value = "OptionROM & UEFI", .key = BOOTIMAGE_PXE_UEFI,	},    \
	{ .value = "OptionROM       ", .key = BOOTIMAGE_PXE,		},    \
	{ .value = "UEFI            ", .key = BOOTIMAGE_UEFI,		},    \
	{ .value = "None            ", .key = BOOTIMAGE_NONE,		}

static const sf_option sf_boot_image_options[] = {
	BOOT_IMAGE_OPTIONS
};

/**************************************************************************
 *
 * MPIO logic
 *
 * MPIO logic is in one place, because it's controlled in two places.
 * One (implicitly) from the adapter menu, and one (explicitly) from
 * the MPIO screen.
 *
 * Because the data is stored in very different formats, there are two
 * functions here, that do similar things
 *
 **************************************************************************
 */
#ifdef BCV_PROTO_ISCSI
static int mpio_device_count ( void ) {
	struct netdev_conf	*conf;
	int			count = 0;

	for_each_iscsi_conf ( conf )
		++count;

	return count;
}
#endif

static void write_adapter_lines ( void )
{
	struct netdev_conf	*conf;
	struct net_device	*net_dev;
	struct pci_device	*pci_dev;
	sf_widget		*line;
	int			boot_image;

	for_each_conf ( conf ) {
		/* Only update primary NICs. */
		if (conf->pf_index != 0)
			continue;

		net_dev = conf->netdev;
		pci_dev =  container_of ( net_dev->dev,
					  struct pci_device, dev );
		line = &conf->main_entry->widget;
		boot_image = conf->_bootimage->index;

		free ( line->text );
		asprintf ( &line->text,
			   "%02x-%02x-%02x-%02x-%02x-%02x  %02x:%02x.%1d  %s",
			   net_dev->ll_addr[0], net_dev->ll_addr[1],
			   net_dev->ll_addr[2], net_dev->ll_addr[3],
			   net_dev->ll_addr[4], net_dev->ll_addr[5],
			   pci_dev->bus, PCI_SLOT ( pci_dev->devfn ),
			   PCI_FUNC ( pci_dev->devfn ),
			   sf_boot_image_options[boot_image].value );

		sf_widget_draw(line);
	}
}

static void sf_check_error ( struct net_device *netdev, int rc, const char *msg1,
			     const char *msg2 )
{
	const char *lines[] = { msg1, "", NULL };
	char *text;

	if ( rc == 0 )
		return;

	asprintf ( &text, " %s%s%s (0x%x) ",
		   ( netdev != NULL ) ? netdev->name : "",
		   ( netdev != NULL ) ? ": " : "",
		   msg2, rc );
	lines[2] = text;

	sf_dialog ( NULL, "Error", SFW_DIALOG_OK, -1, -1,
		    ( const char ** ) lines, ARRAY_SIZE ( lines ) );
	free ( text );
}

static int
nvo_check ( struct net_device *netdev, const char *msg )
{
	int rc = sf_nvo_checks_netdev ( netdev );

	sf_check_error ( netdev, rc, msg, sc_err_msgs [ -rc ] );
	return rc;
}

static int sf_save_all ( struct netdev_conf *conf, sf_widget **widget )
{
	int rc, pos;

	for ( pos = 0; pos < NUM_SCREENS; ++pos ) {
		if (!conf->screens[pos])
			continue;

		if ( ( rc = sf_widget_save ( conf->screens[pos], widget ) ) < 0 )
			return rc;
	}

#ifdef BCV_PROTO_ISCSI
	return sf_widget_save ( &conf->mpio_row->widget, widget);
#else
	return 0;
#endif
}

/* Save all of the widget state to the nvo backing object. Sets
 * conf->dirty for those devices that changed, and returns a
 * boolean indicating whether dirty is now set on any conf.
 */
static int sf_save_check ( struct sf_widget *this __unused ) {
	struct netdev_conf	*conf;
	struct nvo_block	*nvo = NULL;
	void			*original = NULL;
	sf_widget		*widget;
	int			dirty = 0;
	int			rc;

	if (reboot_needed)
		return RC_QUIT;

	for_each_conf ( conf ) {
		if (!conf_has_link_control(conf) || !conf->settings[BOOTROM])
			continue;

		nvo = container_of(conf->settings[BOOTROM], struct nvo_block, settings);
		original = malloc(nvo->total_len);
		if (!original)
			return RC_QUIT;
		memcpy(original, nvo->data, nvo->total_len);

		/* Attempt to write all the settings to EEPROM. If any of these
		 * fail it's because we're out of space, there's no point in
		 * continuing */
		if ( ( rc = sf_save_all ( conf, &widget ) ) < 0 ) {
			const char *msg[] = { "Unable to save setting!", "", "" };
			char title[40];

			snprintf ( title, sizeof ( title ) - 1, " Error: %s ",
				   widget->setting->description );
			msg[2] = strerror ( rc );
			sf_dialog ( NULL, title, SFW_DIALOG_OK, -1, -1,
				    msg, ARRAY_SIZE ( msg ) );

			/* Allow the parent window to quit */
			free ( original );
			return RC_QUIT;
		}

		/* Check for any differences */
		if ((memcmp(original, nvo->data, nvo->total_len) != 0)) {
			/* Setting this here (before we've finished validating
			 * the other adapters) means that if there's a
			 * subsequent validation error on a different adapter,
			 * then this adapter will still be marked as dirty
			 * even if the settings are manually reverted.
			 */
			conf->dirty = 1;
		}
		free(original);
		original = NULL;

		if (conf->settings[DYNAMIC_CONFIG])
			conf->dirty |= dyncfg_changed(conf->settings[DYNAMIC_CONFIG]);

		dirty |= conf->dirty;

		/* dirty may be set from a previous invocation */
		if ( conf->dirty ) {
			rc = nvo_check ( conf->netdev, "Saving cancelled: "
					 "Please correct configuration "
					 "errors: ");
			if ( rc != 0 )
				return RC_INVAL;
		}
	}

	if ( dirty )
		/* Pop up the exit screen */
		return sf_widget_run ( exit_scr, 0 );

	return RC_QUIT;
}

static void sf_reset_settings ( struct netdev_conf *conf )
{
	uint8_t boot_image = BOOTIMAGE_PXE_UEFI;
	int rc;

	assert ( conf );

	rc = clear_settings(conf->settings[DYNAMIC_CONFIG]);
	sf_check_error(conf->netdev, rc, "", "Failed to wipe dynamic config");

	clear_settings(conf->settings[BOOTROM]);

	/* Reset the boot image setting to both OptionROM & UEFI. */
	conf->_bootimage->index = 0;
	rc = store_setting(conf->settings[BOOTROM], &boot_image_setting,
			   &boot_image, sizeof(boot_image));
	sf_check_error ( conf->netdev, rc, "",
			 "Failed to save boot image setting" );
}

static void sf_write_nvo ( struct netdev_conf *conf )
{
	struct nvo_block *nvo;
	int rc;

	assert ( conf );
	nvo = container_of ( conf->settings[BOOTROM], struct nvo_block, settings );

	/* Save the settings to nvram */
	nvo_set_holdoff ( nvo, 0 );
	nvo_set_holdoff ( nvo, 1 );

	if ( conf->settings[DYNAMIC_CONFIG] ) {
		rc = dyncfg_flush ( conf->settings[DYNAMIC_CONFIG] );
		sf_check_error ( conf->netdev, rc, "",
				 "Failed to save dynamic config" );
	}

	/* Assume config has changed and reboot on exit */
	reboot_needed = 1;
}

/**
 * Button callback that sets adapter settings to default values
 * and leaves the form
 *
 * @v this	The widget
 *
 * @ret		RC_COMPLETE
 */
static int sf_button_cb_adapter_defaults ( sf_widget *this ) {
	struct netdev_conf	*conf = this->aux_data;
	struct pci_device	*pci_dev = sf_netdev_pcidev(conf->netdev);
	struct netdev_conf	*pf_conf;
	static const char *lines[] = { "Adapter configuration will be reset "
				       "immediately, ",
				       "and the adapter will do a cold reboot.",
				       NULL, "Continue?" };
	uint8_t this_pci_bus = pci_dev->bus;
	unsigned this_pci_slot = PCI_SLOT(pci_dev->devfn);

	if ( sf_dialog ( NULL, "Reset to defaults",
			 SFW_DIALOG_OKCANCEL, -1, -1,
			 lines, ARRAY_SIZE ( lines ) ) != 0 )
		return 0;

	/* Clear the screen now, since the code below is pretty slow and
	 * the user might think the menu is stuck. It also means debug
	 * messages are readable.
	 */
	curs_set ( 1 );
	endwin ( );

	/* Reset all devices for this adapter. */
	for_each_settings_conf( pf_conf ) {
		pf_conf->dirty = 0;

		if (!conf_has_link_control(pf_conf))
			continue;

		pci_dev = sf_netdev_pcidev(pf_conf->netdev);
		if (pci_dev->bus != this_pci_bus ||
		    PCI_SLOT(pci_dev->devfn) != this_pci_slot)
			continue;

		sf_reset_settings ( pf_conf );
		sf_write_nvo ( pf_conf );
	}

	/* Exit from the adapter window. */
	return RC_QUIT;
}


/**************************************************************************
 *
 * LED blinking
 *
 * Registers a process that blinks the led on and off every 1s, until
 * the user exits the dialog
 *
 **************************************************************************
 */

struct blinker {
	struct refcnt refcnt;
	struct process process;
	struct net_device *netdev;
	struct sf_widget *dialog;
	unsigned long tick;
	enum sf_flash_state state;
};


static void blink ( struct process *process )
{
	struct blinker *blinker = container_of ( process, struct blinker,
						 process );

	if ( blinker->state == FLASH_NONE ||
	     currticks() - blinker->tick > ticks_per_sec() ) {
		blinker->tick = currticks();
		blinker->state = ( blinker->state == FLASH_BLINK_ON ?
				   FLASH_BLINK_OFF : FLASH_BLINK_ON );
		sf_netdev_flash_leds ( blinker->netdev, blinker->state );

		/* Redraw the status line */
		if ( blinker->dialog != NULL ) {
			sf_widget *status = next_widget ( next_widget ( blinker->dialog ) );
			const char *value;

			value = ( blinker->state == FLASH_BLINK_ON ?
				  "               ON              " :
				  "               off             " );

			free ( status->text );
			status->text = strdup ( value );
			sf_widget_draw ( status );
		}
	}
}

static int sf_adapter_cb_vlan_enabled ( sf_widget *this ) {
	set_widgets_selectable ( next_widget ( this ), 1,
				 (this->selectable & SF_CHECKBOX ( this )->checked));

	return 0;
}

static int sf_adapter_cb_flash_leds ( sf_widget *this ) {
	struct netdev_conf *conf = this->aux_data;
	struct blinker *blinker;

	/* Create a process to blink the LED's */
	blinker = zalloc ( sizeof ( *blinker ) );
	if ( blinker == NULL )
		return -ENOMEM;

	/* A PF without link access is not allowed to blink the LEDs, so use
	 * the PF that has link control.
	 */
	if (!conf_has_link_control(conf))
		conf = conf->link_ctrl_conf;
	blinker->netdev = conf->netdev;
	blinker->tick = currticks();
	blinker->state = FLASH_NONE;
	process_init ( &blinker->process, &blink, &blinker->refcnt );
	ref_put ( &blinker->refcnt );

	/* Create the window */
	static const char *lines[] = { "Press ESC to cancel blinking...",
				       NULL, "" };

	sf_dialog ( &blinker->dialog, "Flash Adapter LED", SFW_DIALOG_MSG, 40,
		    ARRAY_SIZE ( lines ) + 3, lines, ARRAY_SIZE ( lines ) );

	process_del ( &blinker->process );
	blinker = NULL;

	sf_netdev_flash_leds ( conf->netdev, FLASH_NONE );
	return 0;
}

/**************************************************************************
 *
 * Main adapter menu
 *
 * Contains a list of adapters, each adapter has one button which displays
 * the adapter menu. At the bottom of the screen there's some additional
 * information about the selected adapter.
 *
 **************************************************************************
 */

/* Read the VPD settings */
static void get_vpd_data(struct netdev_conf *conf)
{
	int rc;

	rc = sf_netdev_vpd(conf->netdev, VPD_ID, 0, (uint8_t *) conf->vpd_id,
			   sizeof(conf->vpd_id) - 1);
	conf->vpd_id[rc > 0 ? rc : 0] = '\0';

	rc = sf_netdev_vpd(conf->netdev, VPD_RO, SF_VPD_KEYWORD('V', '0'),
			   (uint8_t *) conf->vpd_fw_ver,
			   sizeof(conf->vpd_fw_ver) - 1);
	conf->vpd_fw_ver[rc > 0 ? rc : 0] = '\0';
}

static void sf_adapter_cb_adapter_enter ( sf_widget *this ) {
	struct netdev_conf *conf = this->aux_data;

	sf_widget *w = main_scr_summary;

	if ( conf->vpd_id[0] != '\0' ) {
		sf_info_change ( SF_INFO ( w ), "Controller:      ",
				 "%s", conf->vpd_id );
		w = next_widget ( w );
	}

	/* Firmware */
	if ( conf->vpd_fw_ver[0] != '\0' ) {
		sf_info_change ( SF_INFO ( w ), "Firmware:        ", "%s",
				 conf->vpd_fw_ver );
		w = next_widget ( w );
	}

	/* And wipe the remaining boxes */
	for ( ; w->type == SFW_INFO; w = next_widget ( w ) )
		sf_info_change ( SF_INFO ( w ), NULL, NULL );

	for ( w = main_scr_summary; w->type == SFW_INFO; w = next_widget ( w ) )
		sf_widget_draw ( w );
}

static sf_widget *prepare_main_scr ( void ) {
	sf_widget		*list, *window, *w, *container;
	struct netdev_conf	*conf;

	static const struct sf_widget_brief sf_main_conf_screen[] = {
		SF_BR_WINDOW ( 0, " Select Adapter " ),
		SF_BR_TEXT ( "  Base MAC address   PCI      Boot image" ),
		SF_BR_NEWLINE ( 1 ),
		SF_BR_LIST ( 5, SFC_INIT_COL ),
		SF_BR_NEWLINE ( 3 ),
		SF_BR_INFO ( NULL ),  /* Controller */
		SF_BR_INFO ( NULL ),  /* Firmware */
		SF_BR_INFO ( NULL ),  /* Spare */
	};

	static const struct sf_widget_brief sf_main_conf_screen_module[] = {
		SF_BR_CAPTION ( "|", 0, 0 ),
		SF_BR_BUTTON ( NULL, 2, 0, sf_button_cb, MENU,
			       "The list shows every Solarflare adapter "
			       "currently available in this host. Use the "
			       "arrow keys to move between them, and press "
			       "Return to modify the configuration of the "
			       "selected adapter.\n\nTo select the boot order "
			       "of these adapters use the Bios setup screen. "
			       "Refer to your platform BIOS documentation for "
			       "more information." ),
		SF_BR_CAPTION ( " | ", 70, 1 ),
	};

	window = prepare_conf_widgets ( NULL, SFC_INIT_COL, NULL,
					sf_main_conf_screen,
					ARRAY_SIZE ( sf_main_conf_screen ),
					NULL, NULL );

	for ( w = window; w->type != SFW_LIST; w = next_widget ( w ) )
		;
	container = SF_LIST ( w )->child;
	/* Add an exit constraint */
	container->action = sf_save_check;

	for ( ; w->type != SFW_INFO; w = next_widget ( w ) )
		;
	main_scr_summary = w;

	/* Place the list entries at the top */
	br_row = 0;
	for_each_conf ( conf ) {
		sf_widget	*w;

		/* Only list primary NICs here. */
		if (conf->pf_index != 0)
			continue;

		list = prepare_conf_widgets (
			container, 0, conf->settings,
			sf_main_conf_screen_module,
			ARRAY_SIZE ( sf_main_conf_screen_module ), NULL, conf );

		w = next_widget ( list );
		conf->main_entry = SF_BUTTON ( w );
		w->enter = &sf_adapter_cb_adapter_enter;
		w->carret = SF_CARRET_NONE;
	}

	return window;
}


/**************************************************************************
 *
 * Adapter menu
 *
 * A popup menu that allows the user to configure the boot image, and
 * provides buttons to open further popups for iscsi, global and PF options.
 *
 **************************************************************************/
#ifdef BCV_PROTO_ISCSI
#define BOOT_TYPES	\
	{ .value = "PXE     ", .key = BOOTTYPE_PXE, },		\
	{ .value = "iSCSI   ", .key = BOOTTYPE_ISCSI, },	\
	{ .value = "Disabled", .key = BOOTTYPE_DISABLED, }
#define BOOT_TYPE_COUNT	3
#else
#define BOOT_TYPES	\
	{ .value = "PXE     ", .key = BOOTTYPE_PXE, },		\
	{ .value = "Disabled", .key = BOOTTYPE_DISABLED, }
#define BOOT_TYPE_COUNT	2
#endif

static const sf_option boot_type_options[] = {
	BOOT_TYPES
};

static int fetch_pf_count_setting(struct settings *settings[])
{
	long value;
	int rc = fetch_int_setting(settings[DYNAMIC_CONFIG],
				   &pf_count_setting, &value);

	/* If the ROM data is corrupted/missing we use the default value. */
	if (rc < 0 || value == 0)
		return 1;
	return (int) value;
}

/* Populate the text for one PF in the list. */
static void sf_pf_set_row(sf_widget *w, struct netdev_conf *conf)
{
	struct net_device *net_dev = conf->netdev;
	struct pci_device *pci_dev = container_of(net_dev->dev,
						  struct pci_device, dev);
	sf_combo *boottype = conf->boottype;

	if (!conf_has_link_control(conf)) {
		boottype = conf->link_ctrl_conf->boottype;
	}

	if (w->text)
		free(w->text);

	asprintf(&w->text,
		 "%2d %02x-%02x-%02x-%02x-%02x-%02x  %02x:%02x.%1d  %15s",
		 conf->pf_index,
		 net_dev->ll_addr[0], net_dev->ll_addr[1],
		 net_dev->ll_addr[2], net_dev->ll_addr[3],
		 net_dev->ll_addr[4], net_dev->ll_addr[5],
		 pci_dev->bus, PCI_SLOT(pci_dev->devfn),
		 PCI_FUNC(pci_dev->devfn),
		 (boottype ?
		  boot_type_options[boottype->index].value : "")
		 );
}

static void refresh_pf_list ( struct netdev_conf *conf )
{
	sf_widget		*w;

	/* Write all rows. */
	for_each_screen_widget(w, conf, MENU) {
		struct netdev_conf *pf_conf = w->aux_data;

		if (w->aux_data2 != PF)
			continue;

		sf_pf_set_row(w, pf_conf);
		sf_widget_draw(w);
	}
}

static void sf_menu_cb_enter ( sf_widget *this ) {
	struct netdev_conf	*conf = this->aux_data;
	sf_widget		*w;

	for_each_widget(w, this) {
		/* Disable the "Reset to Defaults" option if the adapter is
		 * read-only.
		 */
		if (!conf->writable_settings &&
		    (w->action == sf_button_cb_adapter_defaults)) {
			w->selectable = FALSE;
			break;
		}
	}

	refresh_pf_list(conf);
}

static void sf_menu_cb_exit(sf_widget *this __unused)
{
	write_adapter_lines();
}

/* Prepare adapter menu */
static sf_widget *prepare_menu_scr ( struct netdev_conf *conf ) {
	struct netdev_conf *pf_conf;
	sf_widget *window;
	int pf_count = 0;

	static const struct sf_widget_brief sf_menu_conf_title[] = {
		SF_BR_WINDOW ( 0, " Adapter Menu " )
	};
	static const struct sf_widget_brief sf_menu_conf_ro_title[] = {
		SF_BR_WINDOW ( 0, " Adapter Menu (Read-only) " )
	};

	static const struct sf_widget_brief sf_menu_conf_screen[] = {
		SF_BR_BUTTON ( "Global adapter Options ->",
			       SFC_INIT_COL, 1, sf_button_cb, GLOBAL, NULL ),
		SF_BR_BUTTON ( "Reset to Defaults ->",
			       SFC_INIT_COL, 1,
			       sf_button_cb_adapter_defaults, 0,
			       "Reset the saved configuration state to "
			       "defaults.\n\nWarning: This can not be "
			       " reversed" ),
		SF_BR_NEWLINE ( 1 ),
		SF_BR_TEXT ( "  PF MAC                PCI      Boot type" ),
		SF_BR_NEWLINE ( 1 ),
	};

	static const struct sf_widget_brief sf_per_pf_conf[] = {
		SF_BR_BUTTON(" ", SFC_INIT_COL+2, 1, sf_button_cb, PF,
			     "The list shows every physical function "
			     "currently available on this adapter. Use the "
			     "arrow keys to move between them, and press "
			     "Return to modify the configuration of the "
			     "selected PF.\n"),
	};

	const struct sf_widget_brief *title = sf_menu_conf_title;

	if ( !conf->writable_settings )
		title = sf_menu_conf_ro_title;
	window = prepare_conf_widgets ( NULL, SFC_INIT_COL, conf->settings,
					title, 1, NULL, conf );

	prepare_conf_widgets ( window, SFC_INIT_COL, conf->settings,
			       sf_menu_conf_screen,
			       ARRAY_SIZE ( sf_menu_conf_screen ), NULL,
			       conf );

	/* Count the number of PFs on this adapter.
	 * fetch_pf_count_setting() gives us half of this, rounded down.
	 */
	for_each_conf(pf_conf) {
		if (pf_conf->nicnum < conf->nicnum)
			continue;
		if ((pf_conf != conf) && (pf_conf->pf_index == 0))
			break;
		pf_count++;
	}

	/* Add an entry for each PF on this adapter. */
	for_each_conf(pf_conf) {
		/* Only cover this adapter */
		if (pf_conf->nicnum < conf->nicnum ||
		    pf_conf->nicnum >= (conf->nicnum + pf_count))
			continue;

		/* Exit from the per-PF screen will return here. */
		pf_conf->screens[MENU] = window;

		prepare_conf_widgets(window, 0, conf->settings, sf_per_pf_conf,
				     ARRAY_SIZE(sf_per_pf_conf), NULL, pf_conf);
	}
	window->enter = &sf_menu_cb_enter;
	window->exit = &sf_menu_cb_exit;
	SF_WINDOW(window)->scrollable = 1;

	return window;
}

/**************************************************************************
 *
 * PF options
 *
 * Contains options to specify how gPXE interacts with each PF
 *
 **************************************************************************/
#define MSIX_LIMITS	\
	{ .value = "8",		.key = 8,	}, \
	{ .value = "16",	.key = 16,	}, \
	{ .value = "32",	.key = 32,	}, \
	{ .value = "64",	.key = 64,	}, \
	{ .value = "128",	.key = 128,	}, \
	{ .value = "256",	.key = 256,	}, \
	{ .value = "512",	.key = 512,	}, \
	{ .value = "1024",	.key = 1024,	}

#define VF_MSIX_LIMITS	\
	{ .value = "1",		.key = 1,	}, \
	{ .value = "2",		.key = 2,	}, \
	{ .value = "4",		.key = 4,	}, \
	{ .value = "8",		.key = 8,	}, \
	{ .value = "16",	.key = 16,	}, \
	{ .value = "32",	.key = 32,	}


static int sf_boot_type_cb(sf_widget *this)
{
#ifdef BCV_PROTO_ISCSI
	struct netdev_conf	*conf = this->aux_data;
	sf_widget		*w;

	for_each_widget(w, this) {
		if (conf_has_link_control(conf) && w->aux_data2 == ISCSI) {
			w->selectable = sf_conf_is_iscsi(conf);
			sf_widget_draw(w);
			break;
		}
	}
#else
	this = this;
#endif
	return 0;
}

static void sf_pf_enter(struct sf_widget *this)
{
	struct netdev_conf *conf = this->aux_data;
	sf_widget	   *w;
	int		   pf_count = fetch_pf_count_setting(conf->settings);

	for_each_widget(w, this) {
		/* For PFs without link access almost every field is read-only.
		 * This is also the case in multi-PF mode.
		 */
		if ((!conf->writable_settings ||
		     !conf_has_link_control(conf) || (pf_count > 1)) &&
		    (w->action != sf_adapter_cb_flash_leds)) {
				w->selectable = FALSE;
		}
#ifdef BCV_PROTO_ISCSI
		/* Determine if iSCSI is read-only or read-write. */
		if (conf_has_link_control(conf) && w->aux_data2 == ISCSI) {
			w->selectable = sf_conf_is_iscsi(conf);
		}
#endif

		/* For the PFs without link access show the values for the one
		 * that does have link access.
		 */
		if (!conf_has_link_control(conf)) {
			sf_combo *combo;
			struct netdev_conf *la_conf;

			if (w->setting == &boot_type_setting) {
				combo = SF_COMBO(w);
				la_conf = conf->link_ctrl_conf;
				combo->index = la_conf->boottype->index;
			}
		}
	}
}

static void sf_pf_exit(sf_widget *this)
{
	/* Refresh the PF list */
	refresh_pf_list(this->aux_data);
}

static sf_widget *prepare_pf_scr(struct netdev_conf *conf)
{
	sf_widget *window, *w;

	static const sf_option pf_scr_options[] = {
		BOOT_TYPES,
		MSIX_LIMITS,
		VF_MSIX_LIMITS,
	};

#define COMMON_PF_SCR	\
		SF_BR_WINDOW(2, NULL),						\
		SF_BR_EDITBOX("Boot skip delay time:     ",			\
			      4, BOOTROM, &bootskip_delay_setting,		\
			      &sf_editbox_cb_check,				\
			      "Wait for the given interval (in seconds) to "	\
			      "allow the boot to be skipped by pressing "	\
			      "[Esc].\n\nDefault: 5"),				\
		SF_BR_EDITBOX("Banner delay time:        ", 4, BOOTROM,		\
			      &banner_delay_setting, &sf_editbox_cb_check,	\
			      "Wait for the given interval (in seconds) when "	\
			      "displaying the initial 'Press Ctrl-B' "		\
			      "banner.\n\nDefault: 2"),				\
		SF_BR_COMBO("Boot Type:                ", BOOT_TYPE_COUNT,	\
			     BOOTROM, &boot_type_setting, sf_boot_type_cb,	\
			    "Selects which type of boot to perform."),		\
		SF_BR_EDITBOX("Link-up delay time:       ", 4, BOOTROM,		\
			      &linkup_delay_setting, sf_editbox_cb_check,	\
			      "Wait for the given interval (in seconds) "	\
			      "after the link comes up, before initiating "	\
			      "network communications. This might be "		\
			      "required on some networks that make use of "	\
			      "slow Spanning Tree Protocol.\n\nDefault: 5"),	\
		SF_BR_BUTTON("Flash LEDs", 4, 1, sf_adapter_cb_flash_leds, 0,	\
			     "Flash the LEDs on the Solarflare adapter. The "	\
			     "LEDs will continue to flash until the operation "	\
			     "is cancelled by the user."),			\
		SF_BR_NEWLINE(1)

#ifdef BCV_PROTO_ISCSI
	static const struct sf_widget_brief sf_conf_iscsi_option[] = {
		SF_BR_BUTTON("iSCSI Options ->         ", SFC_INIT_COL, 1,
			     sf_button_cb, ISCSI, NULL)
	};
#endif

	static const struct sf_widget_brief sf_conf_siena_pf_scr[] = {
		COMMON_PF_SCR,
		SF_BR_COMBO("MSI-X interrupt limit:    ",
			    8, BOOTROM, &msix_limit_setting, NULL, NULL),
		SF_BR_CHECKBOX ( "SR-IOV enabled",
				 BOOTROM, &sriov_setting, NULL, NULL ),
		SF_BR_EDITBOX("VF Count:                 ",
			      4, BOOTROM, &vf_count_setting, NULL, NULL),
		SF_BR_COMBO("VF MSI-X interrupt limit: ",
			    6, BOOTROM, &siena_vf_msix_limit_setting, NULL,
			    NULL),
	};
	static const struct sf_widget_brief sf_conf_hunt_pf_scr[] = {
		COMMON_PF_SCR,
		SF_BR_COMBO("MSI-X interrupt limit:      ", 8,
			    DYNAMIC_CONFIG, &msix_limit_setting, NULL, NULL),
		SF_BR_EDITBOX("VF Count:                 ", 4,
			      DYNAMIC_CONFIG, &vf_count_setting, NULL, NULL),
		SF_BR_COMBO("VF MSI-X interrupt limit: ", 6,
			    DYNAMIC_CONFIG, &hunt_vf_msix_limit_setting,
			    NULL, NULL),
		SF_BR_CHECKBOX ("VLAN enabled", DYNAMIC_CONFIG,
				&vlan_setting, sf_adapter_cb_vlan_enabled,
				"Enable VLAN tagging on this port."
				"\n\nDefault: disabled"),
		SF_BR_EDITBOX ("VLAN Tag:                 ", 5,
			       DYNAMIC_CONFIG, &vlan_tag_setting,
			       sf_editbox_cb_check,
			       "VLAN tag to use on this PF. Valid values are "
			       "between 0 and 4094.\n\nDefault: 0"),
	};
	const struct sf_widget_brief *widget_brief;
	size_t widget_brief_len;

	if (is_ef10_device(sf_netdev_pcidev(conf->netdev)->device)) {
		widget_brief = sf_conf_hunt_pf_scr;
		widget_brief_len = ARRAY_SIZE(sf_conf_hunt_pf_scr);
	} else {
		widget_brief = sf_conf_siena_pf_scr;
		widget_brief_len = ARRAY_SIZE(sf_conf_siena_pf_scr);
	}

	if (sf_netdev_option_supported(conf->netdev, DHCP_EB_VLAN_TAG) &&
	    (fetch_pf_count_setting(conf->settings) == 1)) {
		vlan_setting.type = &setting_type_uint8;
		vlan_tag_setting.type = &setting_type_uint32;
	}

	window = prepare_conf_widgets(NULL, SFC_INIT_COL, conf->settings,
				      widget_brief, widget_brief_len,
				      pf_scr_options, conf);
	window->enter = sf_pf_enter;
	window->exit = sf_pf_exit;
	asprintf(&window->text, " PF%d Options ", conf->pf_index);

	/* For PFs with link access hook up the controls that are displayed
	 * on the adapter menu and on the PFs without link access.
	 */
	for_each_widget(w, window) {
		if (conf_has_link_control(conf)) {
			if (w->setting == &boot_type_setting)
				conf->boottype = SF_COMBO(w);
		}
	}
	return window;
}

#ifdef BCV_PROTO_ISCSI
/**************************************************************************
 *
 * iSCSI menu
 *
 **************************************************************************/
static void sf_iscsi_cb_enter(sf_widget *this)
{
	sf_widget	*w;
	int		i, mpio_count;

	/* Set MPIO menu item visibility */
	mpio_count = mpio_device_count();
	for (w = next_widget(this), i = 0; i < 4; i++, w = next_widget(w)) {
		/* Only show the MPIO screen if there are multiple net_devs */
		if (i == 3 && mpio_count <= 1)
			w->selectable = FALSE;
	}
}

static sf_widget *prepare_iscsi_scr(struct netdev_conf *conf)
{
	sf_widget	*window;

	static const struct sf_widget_brief sf_conf_iscsi[] = {
		SF_BR_WINDOW(2, " iSCSI options "),
		SF_BR_NEWLINE(1),
		SF_BR_BUTTON("iSCSI Initiator ->",
			     SFC_INIT_COL, 1, sf_button_cb, ISCSI_INIT,
			     NULL),
		SF_BR_BUTTON("iSCSI Target    ->",
			     SFC_INIT_COL, 1, sf_button_cb, ISCSI_TARGET,
			     NULL),
		SF_BR_BUTTON("iSCSI CHAP      ->",
			     SFC_INIT_COL, 1, sf_button_cb, ISCSI_CHAP,
			     NULL),
		SF_BR_BUTTON("iSCSI MPIO      ->",
			     SFC_INIT_COL, 1, sf_button_cb, _ISCSI_MPIO,
			     NULL),
	};

	window = prepare_conf_widgets(NULL, SFC_INIT_COL, conf->settings,
				      sf_conf_iscsi,
				      ARRAY_SIZE(sf_conf_iscsi), NULL, conf);
	window->enter = &sf_iscsi_cb_enter;

	return window;
};


/**************************************************************************
 *
 * iSCSI target menu
 *
 * Contains a "Use DHCP for iSCSI target" checkbox, and
 * four editboxes (name/address/lun/port), which are only
 * selectable when the checkbox is cleared
 *
 **************************************************************************
 */

static int sf_iscsi_target_cb_dhcp ( sf_widget *this )
{
	set_widgets_selectable ( next_widget ( this ), 4,
				 !SF_CHECKBOX ( this )->checked );

	return 0;
}

static sf_widget *prepare_iscsi_target_scr ( struct netdev_conf *conf ) {
	sf_widget	*list, *w;
	
	static const struct sf_widget_brief sf_conf_iscsi_target[] = {
		SF_BR_WINDOW ( 2, " iSCSI Target settings " ),
		SF_BR_CHECKBOX ( "Use DHCP for Target",
				 BOOTROM, &target_dhcp_setting, sf_iscsi_target_cb_dhcp,
				 "Issue a DHCP request for the runtime iSCSI "
				 "Target configuration, instead of using the "
				 "saved settings." ),
		SF_BR_EDITBOX ( "  Target IQN:          ",
				SF_TARGET_IQN_LEN_MAX + 1, BOOTROM, &target_name_setting,
				NULL,
				"Specifies the IQN to be used by the Target. "
				"The format of this is usually IQN or EUI: "
				"refer to your iSCSI Target documentation "
				"for details of how to configure this setting." ),
		SF_BR_EDITBOX ( "  Target server:       ",
				/* Truncation here is coped with in sf_brwidgets */
				( unsigned char ) (SF_TARGET_SERVER_LEN_MAX + 1),
				BOOTROM, &server_name_setting,
				NULL,
				"The DNS name or IPv4 address of the target "
				"server." ),
		SF_BR_EDITBOX ( "  Target port:         ",
				6, BOOTROM, &port_setting, sf_editbox_cb_check,
				"Connection port number.\n\n"
				"Default: 3260" ),
		SF_BR_EDITBOX ( "  Target LUN:          ",
				4, BOOTROM, &lun_setting, sf_editbox_cb_check,
				"Logical unit number (LUN) of the iSCSI Target."
				"\n\nDefault: 0.\nRange: 0-255\n" ),
		SF_BR_NEWLINE ( 1 ),
		SF_BR_EDITBOX ( "LUN busy retries:      ",
				4, BOOTROM, &lun_retries_setting, sf_editbox_cb_check,
				"Number of times the initiator will attempt to "
				"connect to the iSCSI target local unit "
				"containing the image before giving up.\n\n"
				"Default: 2\nRange: 0-255" ),
	};

	static const sf_option iscsi_target_options[] = {
		{
			.value = "No  ",
			.key = 0,
		},
		{
			.value = "Yes ",
			.key = 1,
		},
	};

	list = prepare_conf_widgets (
		NULL, SFC_INIT_COL, conf->settings, sf_conf_iscsi_target,
		ARRAY_SIZE ( sf_conf_iscsi_target ), iscsi_target_options,
		conf );

	/* Hook up publically accessible controls */
	for ( w = list; w->setting != &target_dhcp_setting; w = next_widget ( w ) )
		;
	conf->target_dhcp = SF_CHECKBOX ( w );
	conf->target_name_edt = SF_EDITBOX ( w = next_widget ( w ) );
	conf->target_server_edt = SF_EDITBOX ( w = next_widget ( w ) );
	conf->target_port_edt = SF_EDITBOX ( w = next_widget ( w ) );
	conf->target_lun_edt = SF_EDITBOX ( w = next_widget ( w ) );

	return list;
}

/**************************************************************************
 *
 * iSCSI initiator menu
 *
 * Contains a "Use DHCP for iSCSI initiator" checkbox, and
 * four editboxes (address/subnet/gateway/dns), which are only
 * selectable when the checkbox is cleared. Also contains
 * an initiator IQN editbox, and a checkbox to select which takes
 * preference, the editbox, or dhcp option 43.203
 *
 **************************************************************************
 */

static int sf_iscsi_initiator_cb_dhcp ( sf_widget *this )
{
	set_widgets_selectable ( next_widget ( this ), 4,
				 !SF_CHECKBOX ( this )->checked );

	return 0;
}

static int sf_iscsi_initiator_cb_iqn_dhcp ( sf_widget *this )
{
	set_widgets_selectable ( next_widget ( this ), 1,
				 !SF_CHECKBOX ( this )->checked );

	return 0;
}

static sf_widget *prepare_iscsi_initiator_scr ( struct netdev_conf *conf ) {
	sf_widget	*list, *w;
	
	static const struct sf_widget_brief sf_conf_iscsi_initiator[] = {
		SF_BR_WINDOW ( 2, " iSCSI initiator settings " ),
		SF_BR_CHECKBOX ( "Use DHCP for Initiator",
				 BOOTROM, &initiator_dhcp_setting,
				 sf_iscsi_initiator_cb_dhcp,
				 "Issue a DHCP request to obtain the runtime "
				 "IP and DNS configuration, instead of using "
				 "the saved settings." ),
		SF_BR_EDITBOX ( "  IPv4 address:        ",
				16,BOOTROM,  &ip_setting, sf_editbox_cb_check,
				"IPv4 address of the initiator adapter. "
				"Enter the address in 'a.b.c.d' format.\n\n"
				"For example: 192.0.0.1" ),
		SF_BR_EDITBOX ( "  IPv4 subnet mask:    ",
				16, BOOTROM, &netmask_setting, sf_editbox_cb_check,
				"IP address subnet mask. Enter the address in "
				"'a.b.c.d' format.\n\n"
				"For example: 255.255.255.0" ),
		SF_BR_EDITBOX ( "  Default gateway:     ",
				16, BOOTROM, &gateway_setting, sf_editbox_cb_check,
				"Network gateway address. A gateway address "
				"may be required if the target is on another "
				"network. Enter the address in 'a.b.c.d' "
				"format.\n\nFor example: 192.168.0.1" ),
		SF_BR_EDITBOX ( "  Primary DNS:         ",
				16, BOOTROM,&dns_setting, sf_editbox_cb_check,
				"Address of a Primary DNS server. Enter the "
				"address in 'a.b.c.d' format.\n\n"
				"For example: 192.168.0.1" ),
		SF_BR_NEWLINE ( 1 ),
		SF_BR_CHECKBOX ( "Use DHCP for initiator IQN",
				 BOOTROM,&initiator_iqn_dhcp_setting,
				 sf_iscsi_initiator_cb_iqn_dhcp,
				 "Use the DHCP offered initiator IQN in "
				 "preference to the current setting. If the "
				 "DHCP offer does not contain an initiator IQN, "
				 "then the current setting is used instead." ),
		SF_BR_EDITBOX ( "  Initiator IQN:       ",
				SF_INITIATOR_LEN_MAX + 1, BOOTROM, &initiator_iqn_setting, NULL,
				"Name of the iSCSI Initiator. The initiator name "
				"must be known to the iSCSI Target.\n\n"
				"For example: iqn.2009-01.com.solarflare" ),
		SF_BR_NEWLINE ( 1 ),
		SF_BR_EDITBOX ( "DHCP Vendor Class ID:  ", 13,
				BOOTROM, &dhcp_vendor_class_setting, NULL,
				"The DHCP vendor class used for ProxyDHCP. Do "
				"not change this setting unless your "
				"application vendor advises you to." ),
	};

	list = prepare_conf_widgets (
		NULL, SFC_INIT_COL, conf->settings, sf_conf_iscsi_initiator,
		ARRAY_SIZE ( sf_conf_iscsi_initiator ), NULL, conf );
	
	/* Hook up publically accessible controls */
	conf->initiator_dhcp = SF_CHECKBOX ( w = next_widget ( list ) );
	conf->initiator_ip = SF_EDITBOX ( w = next_widget ( w ) );
	conf->initiator_subnet = SF_EDITBOX ( w = next_widget ( w ) );
	conf->initiator_gateway = SF_EDITBOX ( w = next_widget ( w ) );

	return list;
}

/**************************************************************************
 *
 * iSCSI CHAP menu
 *
 * Contains two checkboxes, one for chap, one for mutual chap. Each
 * checkbox has two associated editboxes, one for the username,
 * and one for the secret.
 *
 **************************************************************************
 */

static int sf_iscsi_chap_cb_chap ( sf_widget *this ) {
	int		chap = SF_CHECKBOX ( this )->checked;
	int		i;
	sf_widget	*w;
	sf_checkbox	*mutual;

	/* Chap */
	for ( i = 0, w = next_widget ( this ); i < 3; ++i, w = next_widget ( w ) ) {
		w->selectable = chap;
		sf_widget_draw ( w );
	}

	mutual = SF_CHECKBOX ( prev_widget ( w ) );
	if ( chap )
		chap = mutual->checked;
	else
		mutual->checked = 0;
	sf_widget_draw ( prev_widget ( w ) );
	for ( i = 0; i < 2; ++i, w = next_widget ( w ) ) {
		w->selectable = chap;
		sf_widget_draw ( w );
	}

	return 0;
}

static int sf_iscsi_chap_cb_mutual ( sf_widget *this )
{
	sf_widget *w = this;
	unsigned int i;

	for ( i = 0; i < 3; i++ )
		w = prev_widget ( w );

	return sf_iscsi_chap_cb_chap ( w );
}

static sf_widget *prepare_iscsi_chap_scr ( struct netdev_conf *conf ) {

	static const struct sf_widget_brief sf_conf_iscsi_chap[] = {
		SF_BR_WINDOW ( 2, " iSCSI CHAP settings " ),
		SF_BR_CHECKBOX ( "CHAP authentication",
				 BOOTROM, &chap_setting, sf_iscsi_chap_cb_chap,
				 "Enables or Disables the use of CHAP.\n"
				 "If CHAP is employed on the target, specify "
				 "the username and secret." ),
		SF_BR_EDITBOX ( "  User name:        ",
				SF_USERID_LEN_MAX + 1, BOOTROM, &username_setting, NULL,
				"The CHAP user name assocated with this initiator.\n"
				"Length: Maximum of 64 characters." ),
		SF_BR_PASSWORD ( "  Secret:           ",
				 SF_PASSWD_LEN_MAX + 1, BOOTROM, &password_setting,
				 sf_editbox_cb_check_secret,
				 "The CHAP secret associated with this Initiator.\n"
				 "Note: The secret must not contain any '\"' characters.\n" 
				 "Length: Maximum 20, minimum 12 characters." ),
		SF_BR_NEWLINE ( 1 ),
		SF_BR_CHECKBOX ( "Mutual CHAP authentication",
				 BOOTROM, &chap_target_setting, sf_iscsi_chap_cb_mutual,
				 "Enables or Disables the use of Mutual CHAP.\n"
				 "If Mutual CHAP is enabled, specify the target "
				 "username and target secret.\n "),
		SF_BR_EDITBOX ( "  Target user name: ",
				SF_USERID_LEN_MAX + 1, BOOTROM, &reverse_username_setting,
				NULL,
				"The CHAP username associated with the Target.\n"
				"Length: Maximum of 64 characters." ),
		SF_BR_PASSWORD ( "  Target secret:    ",
				 SF_PASSWD_LEN_MAX + 1, BOOTROM, &reverse_password_setting,
				 sf_editbox_cb_check_secret,
				 "The CHAP secret associated with the Target.\n"
				 "Note: The secret must not contain any '\"' characters.\n" 
				 "Length: Maximum 20, minimum 12 characters." ),
	};

	return prepare_conf_widgets (
		NULL, SFC_INIT_COL, conf->settings, sf_conf_iscsi_chap,
		ARRAY_SIZE ( sf_conf_iscsi_chap ), NULL, conf );
}

/**************************************************************************
 *
 * iSCSI MPIO menu
 *
 * Contains a singe list box. The elements of the list box are populated
 * dynamically in sf_menu_cb_iscsi_mpio()
 *
 **************************************************************************
 */

static void sf_mpio_cb_enter ( struct sf_widget *this __unused )
{
	reorder_mpio ( );
}

static int sf_mpio_cmp ( const sf_row *a, const sf_row *b )
{
	struct netdev_conf *conf_a = a->widget.aux_data;
	struct netdev_conf *conf_b = b->widget.aux_data;

	assert ( a->widget.row == b->widget.row );
	return memcmp ( conf_a->netdev->ll_addr,
			conf_b->netdev->ll_addr, 6 );
}

static void reorder_mpio ( void )
{
	struct netdev_conf	*conf;
	int			iscsi;

	for_each_conf ( conf ) {
		iscsi = sf_conf_is_iscsi ( conf );

		conf->mpio_row->widget.selectable = iscsi;
		conf->mpio_row->widget.visible = iscsi;
	}

	sf_row_order ( mpio_scr_list, sf_mpio_cmp );
}

static sf_widget *prepare_iscsi_mpio_scr ( void ) {
	struct netdev_conf	*conf;
	sf_widget		*list, *window, *w;

	static const struct sf_widget_brief sf_mpio_conf_screen_top[] = {
		SF_BR_WINDOW ( 2, " MPIO Priorities " ),
		SF_BR_TEXT ( "     MAC address            PCI" ),
		SF_BR_NEWLINE ( 1 ),
		SF_BR_LIST ( 5, 1 ),
		SF_BR_NEWLINE ( 1 ),
		SF_BR_TEXT ( " Use +/- to change the connection order. " ),
	};

	static const struct sf_widget_brief sf_mpio_conf_screen_module[] = {
		SF_BR_ROW ( NULL, BOOTROM, &mpio_priority_setting,
			    "An iSCSI login to the target will be "
			    "attempted via each iSCSI adapter in turn, "
			    "once the link has come up. Adjust the "
			    "order here, using +/- to move an adapter, "
			    "and up|down to move between adapters.\n\n"
			    "The top most adapter will be used "
			    "first." ),
	};

	window = prepare_conf_widgets (
		NULL, SFC_INIT_COL, NULL, sf_mpio_conf_screen_top,
		ARRAY_SIZE ( sf_mpio_conf_screen_top ), NULL, NULL );
	window->enter = sf_mpio_cb_enter;
	for ( w = window; w->type != SFW_LIST; w = next_widget ( w ) )
		;
	mpio_scr_list = SF_LIST ( w )->child;

	br_row = 0;
	for_each_conf ( conf ) {
		list = prepare_conf_widgets (
			mpio_scr_list, SFC_INIT_COL, conf->settings,
			sf_mpio_conf_screen_module,
			ARRAY_SIZE ( sf_mpio_conf_screen_module ),
			NULL, conf );

		/* Write NIC info */
		struct net_device *netdev = conf->netdev;
		struct pci_device *pci_dev = container_of ( netdev->dev,
							    struct pci_device, dev );
		asprintf ( & list->text,
			   "%02x:%02x:%02x:%02x:%02x:%02x      %02x:%02x.%1d",
			   netdev->ll_addr[0], netdev->ll_addr[1],
			   netdev->ll_addr[2], netdev->ll_addr[3],
			   netdev->ll_addr[4], netdev->ll_addr[5],
			   pci_dev->bus, PCI_SLOT ( pci_dev->devfn ),
			   PCI_FUNC ( pci_dev->devfn ) );

		/* Hook in MPIO row */
		conf->mpio_row = SF_ROW ( list );
	}

	return window;
}
#endif	/* BCV_PROTO_ISCSI */

/**************************************************************************
 *
 * Global adapter options
 *
 * Shows read-only options and can allow setting of the port mode.
 *
 **************************************************************************/

static int sf_port_mode_cb(sf_widget *this)
{
	struct netdev_conf	*conf = this->aux_data;
	sf_combo		*combo = SF_COMBO(this);
	int			idx;

	/* Skip to the next valid port mode.
	 * Index 0 is the default port mode, which is always valid.
	 */
	for (idx = combo->index; idx < combo->varn; idx++) {
		if (!idx || (conf->port_modes & BIT(combo->vars[idx].key)))
			break;
	}
	if (idx == combo->varn)
		combo->index = 0;
	else
		combo->index = idx;
	return 0;
}

static void sf_global_scr_enter(sf_widget *this)
{
	struct netdev_conf	*conf = this->aux_data;
	struct net_device	*netdev = conf->netdev;
	struct sf_widget	*w;

	/* Disable any controls the NIC doesn't support. */
	for (w = next_widget(this); w != this; w = next_widget(w))
		if (!conf->writable_settings ||
		    (w->setting &&
		     !sf_netdev_option_supported(netdev, w->setting->tag)))
			w->selectable = FALSE;
}

static void sf_global_scr_exit(sf_widget *this)
{
	/* Update the list of PFs in case a change in port mode has reset
	 * everything.
	 */
	refresh_pf_list(this->aux_data);
}

static sf_widget *prepare_global_scr(struct netdev_conf *conf)
{
	sf_widget *window, *w;

	static const sf_option global_scr_options[] = {
		{
			.value = "Default             ",
			.key = SFC_PORT_MODE_DEFAULT,
		},
		{
			.value = "[1x40G]             ",
			.key = SFC_PORT_MODE_40G,
		},
		{
			.value = "[1x10G][1x10G]      ",
			.key = SFC_PORT_MODE_10G_10G,
		},
		{
			.value = "[1x40G][1x40G]      ",
			.key = SFC_PORT_MODE_40G_40G,
		},
		{
			.value = "[4x10G]             ",
			.key = SFC_PORT_MODE_4x10G,
		},
		{
			.value = "[2x10G][2x10G]      ",
			.key = SFC_PORT_MODE_10G_10G_10G_10G,
		},
		{
			.value = "[1x100G]            ",
			.key = SFC_PORT_MODE_100G,
		},
		{
			.value = "[1x10/25G][1x10/25G]",
			.key = SFC_PORT_MODE_25G_25G,
		},
		{
			.value = "[1x50G][1x50G]      ",
			.key = SFC_PORT_MODE_50G_50G_Q,
		},
		{
			.value = "[4x10/25G]          ",
			.key = SFC_PORT_MODE_4x25G,
		},
		{
			.value = "[2x10/25G][2x10/25G]",
			.key = SFC_PORT_MODE_25G_25G_25G_25G,
		},
		{
			.value = "[2x50G]             ",
			.key = SFC_PORT_MODE_50G_50G,
		},
		{
			.value = "Auto             ",
			.key = TLV_FIRMWARE_VARIANT_DRIVER_SELECTED,
		},
		{
			.value = "Ultra Low Latency",
			.key = TLV_FIRMWARE_VARIANT_LOW_LATENCY
		},
		{
			.value = "Full Feature     ",
			.key = TLV_FIRMWARE_VARIANT_FULL_FEATURED
		},
		{
			.value = "Packed Stream    ",
			.key = TLV_FIRMWARE_VARIANT_PACKED_STREAM,
		},
		{
			.value = "High TX Rate     ",
			.key = TLV_FIRMWARE_VARIANT_HIGH_TX_RATE,
		},
		{
			.value = "Packed Stream HM1",
			.key = TLV_FIRMWARE_VARIANT_PACKED_STREAM_HASH_MODE_1,
		},
		{
			.value = "Rules Engine     ",
			.key = TLV_FIRMWARE_VARIANT_RULES_ENGINE,
		},
		{
			.value = "DPDK             ",
			.key = TLV_FIRMWARE_VARIANT_DPDK,
		},
		{
			.value = "Layer 3 UDP      ",
			.key = TLV_FIRMWARE_VARIANT_L3XUDP,
		}
	};

#define COMMON_GLOBAL_SCR	\
		SF_BR_WINDOW ( 2, " Global Adapter Options " ),		      \
		SF_BR_BUTTON ( "Rescue Options ->",			      \
			     SFC_INIT_COL, 1, sf_button_cb, RESCUE, NULL ),   \
		SF_BR_BUTTON ( "Use F1 for help on the fields below",	      \
			     SFC_INIT_COL, 1, NULL, 0,			      \
			     "You will need to use sfboot from the sfutils "  \
			     "package to change these settings. See downloads"\
			     " from https://support.solarflare.com.\nYou can "\
			     "reset them to their defaults in the Adapter "   \
			     "Menu.\n" ),				      \
		SF_BR_COMBO ( "Port Mode:                 ",		      \
			      12, DYNAMIC_CONFIG, &port_mode_setting,	      \
			      sf_port_mode_cb,				      \
			      "This selects the port mode for the adapter. "  \
			      "Changing this will reset settings to default, "\
			      "and the machine will need a cold reboot."      \
			    ),						      \
		SF_BR_NEWLINE ( 1 )

	static const struct sf_widget_brief sf_conf_siena_global_scr[] = {
		COMMON_GLOBAL_SCR,
	};
	static const struct sf_widget_brief sf_conf_hunt_global_scr[] = {
		COMMON_GLOBAL_SCR,
		SF_BR_CHECKBOX ( "PFIOV enabled", DYNAMIC_CONFIG,
				 &pfiov_setting, NULL, NULL ),
		SF_BR_COMBO ( "Firmware Variant:          ", 9,
			      DYNAMIC_CONFIG, &firmware_variant_setting,
			      NULL, NULL ),
		SF_BR_CHECKBOX ( "Allow insecure filters", DYNAMIC_CONFIG,
				 &insecure_filters_setting, NULL, NULL ),
	};
	const struct sf_widget_brief *widget_brief;
	size_t widget_brief_len;

	if ( is_ef10_device ( sf_netdev_pcidev ( conf->netdev )->device ) ) {
		widget_brief = sf_conf_hunt_global_scr;
		widget_brief_len = ARRAY_SIZE ( sf_conf_hunt_global_scr );
	} else {
		widget_brief = sf_conf_siena_global_scr;
		widget_brief_len = ARRAY_SIZE ( sf_conf_siena_global_scr );
	}

	if (sf_netdev_option_supported(conf->netdev, DHCP_EB_PORT_MODE) &&
	    (fetch_pf_count_setting(conf->settings) == 1))
		port_mode_setting.type = &setting_type_int32;

	window = prepare_conf_widgets (
		NULL, SFC_INIT_COL, conf->settings,
		widget_brief, widget_brief_len,
		global_scr_options, conf );
	window->enter = sf_global_scr_enter;
	window->exit = sf_global_scr_exit;

	/* Hook externally visible controls */
	for_each_widget(w, window) {
		if (w->setting == &firmware_variant_setting) {
			conf->fw_variant = SF_COMBO(w);
			break;
		}
	}

	return window;
}

/**************************************************************************
 *
 * Rescue options
 *
 * This allows setting of the port mode.
 *
 **************************************************************************/

static int sf_boot_image_cb(sf_widget *this)
{
	struct netdev_conf	*conf = this->aux_data;
	struct pci_device	*pci_dev = sf_netdev_pcidev(conf->netdev);
	uint8_t			this_pci_bus = pci_dev->bus;
	unsigned		this_pci_slot = PCI_SLOT(pci_dev->devfn);
	sf_combo		*combo = SF_COMBO(this);
	struct netdev_conf	*pf_conf;

	/* Update the other port on this adapter as well */
	for_each_conf(pf_conf) {

		if (!conf_has_link_control(pf_conf) ||
		    (pf_conf->port == conf->port))
			continue;

		pci_dev = sf_netdev_pcidev(pf_conf->netdev);
		if (pci_dev->bus != this_pci_bus ||
		    PCI_SLOT(pci_dev->devfn) != this_pci_slot)
			continue;

		pf_conf->_bootimage->index = combo->index;
	}
	return 0;
}

static sf_widget *prepare_rescue_scr(struct netdev_conf *conf)
{
	static const sf_option rescue_scr_options[] = {
		BOOT_IMAGE_OPTIONS,
	};
	static const struct sf_widget_brief sf_conf_rescue_scr[] = {
		SF_BR_WINDOW( 2, " Rescue Menu "),
		SF_BR_NEWLINE(1),
		SF_BR_COMBO ( "Boot Image:        ", 4, BOOTROM,
			      &boot_image_setting, sf_boot_image_cb,
			      "Specifies which firmware image the adapter "
			      "will provide to the host. Only change this "
			      "on instruction from the Solarflare support "
			      "team" ),
	};
	sf_widget *window, *w;

	window = prepare_conf_widgets(NULL, SFC_INIT_COL+2, conf->settings,
				      sf_conf_rescue_scr,
				      ARRAY_SIZE(sf_conf_rescue_scr),
				      rescue_scr_options, conf );

	/* Hook externally visible controls */
	for_each_widget(w, window) {
		if (w->setting == &boot_image_setting) {
			conf->_bootimage = SF_COMBO(w);
		}
	}
	return window;
}

/**************************************************************************
 *
 * Exit screen
 *
 **************************************************************************/
static int sf_button_cb_exit_save ( sf_widget __unused *this )
{
	struct netdev_conf	*conf;

	/* Restore writing */
	for_each_conf ( conf ) {
		if ( conf->writable_settings && conf->dirty )
			sf_write_nvo(conf);
	}

	return RC_QUIT;
}

static int sf_button_cb_exit_discard ( sf_widget __unused *this )
{
	struct netdev_conf	*conf;
	struct nvo_block	*nvo;

	/* Reread the NVO (and leave holdoff enabled) */
	for_each_conf ( conf ) {
		if ( conf->dirty ) {
			nvo = container_of ( conf->settings[BOOTROM], struct nvo_block,
					     settings );
			nvo_load ( nvo );
			if ( conf->settings[DYNAMIC_CONFIG] )
				dyncfg_read ( conf->settings[DYNAMIC_CONFIG] );
			conf->dirty = 0;
		}
	}

	return RC_QUIT;
}

static int sf_button_cb_exit_cancel ( sf_widget __unused *this )
{
	return RC_COMPLETE;
}

static sf_widget *prepare_exit_scr ( void ) {
	static sf_widget *window;
	const int height = 7, width = 34;

	static const struct sf_widget_brief sf_exit_screen[] = {
		SF_BR_BUTTON ( "Save Changes and exit",
			       SFC_INIT_COL, 1, sf_button_cb_exit_save, 0,
			       NULL ),
		SF_BR_BUTTON ( "Discard changes and exit",
			       SFC_INIT_COL, 1, sf_button_cb_exit_discard, 0,
			       NULL ),
		SF_BR_BUTTON ( "Return to setup",
			       SFC_INIT_COL, 1, sf_button_cb_exit_cancel, 0,
			       NULL ),
	};

	/* Create the window manually to adjust the dimensions */
	window = sf_window_init ( NULL,
				  ( SF_WIN_HEIGHT - height ) / 2,
				  ( SF_WIN_WIDTH - width ) / 2,
				  width, height );
	br_row = SFC_INIT_ROW;
	prepare_conf_widgets (
		window, SFC_INIT_COL, NULL, sf_exit_screen,
		ARRAY_SIZE ( sf_exit_screen ), NULL, NULL );

	return window;
}

/**
 * Free all the memory
 */
static void free_all_screens ( void ) {
	struct netdev_conf *conf;
	unsigned int pos;

	if ( all_netdevs == NULL )
		return;

	for_each_conf ( conf ) {
		if (conf->pf_index == 0)
			sf_widgets_list_free ( conf->screens[MENU] );

		for ( pos = MENU+1; pos < NUM_SCREENS; ++pos )
			sf_widgets_list_free ( conf->screens[pos] );
	}

	free ( all_netdevs );
	all_netdevs = NULL;

	sf_widgets_list_free ( main_scr );
	main_scr = NULL;

	sf_widgets_list_free ( exit_scr );
	exit_scr = NULL;

#ifdef BCV_PROTO_ISCSI
	sf_widgets_list_free ( mpio_scr );
	mpio_scr = NULL;
#endif
}

/* So far SFC NICs have a maximum of 4 physical ports. */
#define MAX_PORT	4

/**
 * Prepare main screen, iSCSI boot configuration screen, 
 * and CHAP configuration screen
 */
static int prepare_all_screens ( void ) {
	struct net_device	*netdev = NULL;
	struct netdev_conf	*conf;
	struct netdev_conf	*port_conf[MAX_PORT];
	struct nvo_block	*nvo;
	int			nicnum, rc;

	nicnum = 0;
	for_each_netdev ( netdev )
		++nicnum;

	if (nicnum == 0) {
		printf("Cannot find any devices supported by sfconfig\n");
		printf("Please upgrade to a version newer than: %s\n",
		       PRODUCT_VERSION);
		mdelay(5000);
		return -ENODEV;
	}
	
	all_netdevs = calloc ( sizeof ( struct netdev_conf ),
			       nicnum + 1 );
	if ( all_netdevs == NULL )
		return -ENOMEM;

	/* Populate the fields that are needed in all devices. */
	nicnum = 0;
	for_each_netdev ( netdev ) {
		struct pci_device	*pci_dev = sf_netdev_pcidev(netdev);
		unsigned int		readonly;

		conf = all_netdevs + nicnum;
		conf->netdev = netdev;
		conf->nicnum = nicnum;
		++nicnum;

		rc = netdev_ioctl(netdev, NET_GET_PF_INDEX,
				  (unsigned long) &conf->pf_index);
		if (rc)
			conf->pf_index = PCI_FUNC(pci_dev->devfn);
		rc = netdev_ioctl(netdev, NET_GET_PORT_NUMBER,
			     (unsigned long) &conf->port);
		if (rc || (conf->port >= MAX_PORT))
			return -ENOMEM;

		conf->writable_settings = TRUE;
		/* If the device has no option rom or no nvo settings block,
		 * then create the netdev, but make the settings read-only.
		 */
		if (!sf_netdev_has_option_rom(netdev)) {
			conf->writable_settings = FALSE;
		} else {
			conf->settings[BOOTROM] = netdev_nvo_settings(netdev);
			if (conf->settings[BOOTROM] == NULL)
				conf->writable_settings = FALSE;
		}
		if (sf_netdev_has_dyncfg(netdev)) {
			conf->settings[DYNAMIC_CONFIG] =
				netdev_dyncfg_settings(netdev);
			if (conf->settings[DYNAMIC_CONFIG])
				dyncfg_read(conf->settings[DYNAMIC_CONFIG]);
		}

		/* A PF without link access uses some settings from the PF with
		 * link access (on the same port).
		 */
		if (conf->writable_settings) {
			port_conf[conf->port] = conf;
			conf->link_ctrl_conf = conf;
		} else {
			conf->settings[BOOTROM] =
				port_conf[conf->port]->settings[BOOTROM];
			if (!conf->settings[DYNAMIC_CONFIG])
				conf->settings[DYNAMIC_CONFIG] =
				  port_conf[conf->port]->settings[DYNAMIC_CONFIG];
			conf->link_ctrl_conf = port_conf[conf->port];
		}

		rc = netdev_ioctl(netdev, NET_NIC_IS_READONLY,
				  (unsigned long) &readonly);
		if ((rc == 0) && readonly)
			conf->writable_settings = FALSE;

		DBG("%s@%p: PF %d port %d writable=%d %p %p\n", __FUNCTION__,
		    conf, conf->pf_index, conf->port, conf->writable_settings,
		    conf->settings[BOOTROM], conf->settings[DYNAMIC_CONFIG]);

		/* The PF screen exists for every device. */
		conf->screens[PF] = prepare_pf_scr(conf);
		if (!conf->screens[PF])
			return -ENOMEM;
	}

	/* Populate fields that are only needed in devices with link control */
	for_each_conf(conf) {
		if (!conf_has_link_control(conf))
			continue;

		/* Disable writing to the EEPROM */
		nvo = container_of ( conf->settings[BOOTROM], struct nvo_block,
				     settings );
		nvo_set_holdoff ( nvo, 1 );

		/* Only needed for primary NICs. */
		if (conf->pf_index == 0) {
			get_vpd_data(conf);

			netdev_ioctl(conf->netdev, NET_GET_PORT_MODES,
				     (unsigned long) &conf->port_modes);

			conf->screens[MENU] = prepare_menu_scr(conf);
			if (!conf->screens[MENU])
				return -ENOMEM;
		}
#ifdef BCV_PROTO_ISCSI
		conf->screens[ISCSI] = prepare_iscsi_scr ( conf );
		if ( ! conf->screens[ISCSI] )
			return -ENOMEM;
		conf->screens[ISCSI_TARGET] = prepare_iscsi_target_scr ( conf );
		if ( ! conf->screens[ISCSI_TARGET] )
			return -ENOMEM;
		conf->screens[ISCSI_INIT] = prepare_iscsi_initiator_scr ( conf );
		if ( ! conf->screens[ISCSI_INIT] )
			return -ENOMEM;
		conf->screens[ISCSI_CHAP] = prepare_iscsi_chap_scr ( conf );
		if ( ! conf->screens[ISCSI_CHAP] )
			return -ENOMEM;
#endif
		conf->screens[GLOBAL] = prepare_global_scr ( conf );
		if ( ! conf->screens[GLOBAL] )
			return -ENOMEM;
		conf->screens[RESCUE] = prepare_rescue_scr ( conf );
		if ( ! conf->screens[RESCUE] )
			return -ENOMEM;
	}

	if ( ( exit_scr = prepare_exit_scr ( ) ) == NULL )
		return -ENOMEM;
#ifdef BCV_PROTO_ISCSI
	if ( ( mpio_scr = prepare_iscsi_mpio_scr ( ) ) == NULL )
		return -ENOMEM;
	/* MPIO settings are used on the main screen */
#endif
	if ( ( main_scr = prepare_main_scr ( ) ) == NULL )
		return -ENOMEM;

	/* Update everything */
#ifdef BCV_PROTO_ISCSI
	reorder_mpio () ;
#endif
	write_adapter_lines ( );

	if ( DBG_LOG )
		sleep(4);
	return 0;
}

int sfconfig_run ( void ) {
	struct net_device *netdev;
	int rc;

	reboot_needed = 0;
	if ( ( rc = prepare_all_screens ( ) ) != 0 )
		goto out;

	/* Warn about the first failure - don't keep repeating for
	 * each device */
	for_each_netdev ( netdev ) {
		if ( nvo_check ( netdev, "Warning:" ) != 0 )
			break;
	}

	( void ) sf_widget_run ( main_scr, 0 );

out:
	free_all_screens ( );
	return reboot_needed;
}
