/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */
#ifndef	_GPXE_CLP_H
#define _GPXE_CLP_H

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <gpxe/list.h>
#include <gpxe/pci.h>


#define CLP_MAX_STR_LEN		128
#define CLP_INPUT_BUF_SIZE 256
#define CLP_OUTPUT_BUF_SIZE 4096

#define CLP_COMMAND_COMPLETED          0
#define CLP_COMMAND_SPAWNED            1
#define CLP_COMMAND_PROCESSING_FAILED  2
#define CLP_COMMAND_EXECUTION_FAILED   3

#define CLP_SUPPORT_NO                 (0 << 8)
#define CLP_SUPPORT_BASE               (1 << 8)
#define CLP_SUPPORT_EXTENDED           (2 << 8) /* 2-14 valid, see HSS spec. */
#define CLP_SUPPORT_FULL               (15 << 8)

#define CLP_STATUS_SUCCESS (CLP_COMMAND_COMPLETED | CLP_SUPPORT_BASE)
#define CLP_STRING_SUCCESS "status=0\nstatus_tag=COMMAND COMPLETED\nendoutput\n"

#define CLP_ERROR_STATUS(CODE) (CLP_COMMAND_PROCESSING_FAILED | (CODE << 8))
#define CLP_ERROR_STRING(CODE, STRING) \
                               "status=2\n" \
                               "status_tag=COMMAND PROCESSING FAILED\n" \
                               "error=" #CODE "\n" \
                               "error_tag=" STRING "\n" \
                               "endoutput\n"

#define CLP_STATUS_SYNTAX_ERROR CLP_ERROR_STATUS(252)
#define CLP_STRING_SYNTAX_ERROR CLP_ERROR_STRING(252, "COMMAND SYNTAX ERROR")
#define CLP_STATUS_INVALID_TARGET CLP_ERROR_STATUS(246)
#define CLP_STRING_INVALID_TARGET CLP_ERROR_STRING(246, "INVALID TARGET")
#define CLP_STATUS_UNSUPPORTED CLP_ERROR_STATUS(254)
#define CLP_STRING_UNSUPPORTED CLP_ERROR_STRING(254, "COMMAND NOT SUPPORTED")
#define CLP_STATUS_INTERNAL_ERROR CLP_ERROR_STATUS(255)
#define CLP_STRING_INTERNAL_ERROR CLP_ERROR_STRING(255, "COMMAND ERROR - UNSPECIFIED")

/* CLP format flags */
#define CLP_FORMAT_OEMHP_BINARY 0x1

/* List of CLP-able devices exposed. */
extern struct list_head clp_devices;

struct clp_response {
    char *string;
    int status;
    int flags;
    int length;
};


struct clp_device {
	/* The code keeps track of CLP devices in a linked list. */
	struct list_head list_head;
	/* See interface/smash/clp.c:clp_device_from_netport(). */
	int netport;
	struct pci_device *pci_dev;
	/* Pointer to associated device, which is driver-specific. */
	void *device;
	void (*reset)(struct clp_device *clp_dev,
		      struct clp_response *response);
	int (*get_mac)(struct clp_device *clp_dev, uint8_t *mac);
	void (*set_perm_mac)(struct clp_device *clp_dev,
			     struct clp_response *response, const uint8_t *mac);
	void (*set_boot)(struct clp_device *clp_dev,
			 struct clp_response *response, int enable);
};

void register_clp_device(struct clp_device *dev);
void unregister_clp_device(struct clp_device *dev);
struct clp_device *clp_device_from_netport(int netport);
void clp_parse(struct clp_response *response, char *input);


/* HP-specific definitions for CLP command parsing */
enum hp_lf_args {
	HP_LF_PROTOCOL = 0,
	HP_LF_PRIORITY,
	HP_LF_MAC,
	HP_LF_VNI,
	HP_LF_MINBW,
	HP_LF_MAXBW,
	HP_LF_BOOT,
	HP_LF_ENABLE,
	HP_LF_SRIOV,
};

#endif
